#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -------------------------------------------------------------------------------------
# GeoCLUSTER, data tool and app that runs visuals on Closed-Loop Geothermal Data
# -------------------------------------------------------------------------------------

# --------------------
# Libraries.
# --------------------

# web app and interactive graphics libraries 
import dash
from dash.dependencies import Input, Output, State
from dash import Dash, dcc, html, ctx
import dash_daq as daq                  # Adds more data acquisition (DAQ) and controls to dash callbacks 
import dash_bootstrap_components as dbc # Adds bootstrap components for more web themes and templates
from dash.exceptions import PreventUpdate

# sourced scripts
from write2excel import write_excelsheet
from sliders import *
from dropdowns import *
from text import *
from tables import generate_summary_table
from plots import * # u_sCO2, u_H2O, c_sCO2, c_H2O

# -----------------------------------------------------------------------------
# Create dash app with CSS styles and HTML components.
#
#   layout:      Describes the HTML layout of the app.
#   callbacks:   Interactivity of the app.
#
# -----------------------------------------------------------------------------

app = dash.Dash(__name__, external_stylesheets=[dbc.themes.BOOTSTRAP], url_base_pathname="/GeoCLUSTER/")

# -----------------------------------------------
# Globl styles, configurations, and colors.
# -----------------------------------------------

tabs = ["about-tab", "energy-time-tab", "energy-tab", "econ-time-tab", "summary-tab"]


plotly_config = {'displaylogo': False,
                'modeBarButtonsToRemove': ['autoScale', 'resetScale'], # High-level: zoom, pan, select, zoomIn, zoomOut, autoScale, resetScale
                'toImageButtonOptions': {
                    'format': 'png', # one of png, svg, jpeg, webp
                    'filename': 'custom_image',
                    'height': None,
                    'width': None,
                    'scale': 6 # Multiply title/legend/axis/canvas sizes by this factor
                        }
                  }


darkergrey = "#383838"
lightbrown = '#ede6dd'
dropdown_guidance_style = {'position': 'relative',
                            'marginTop': '28px',
                            # 'marginTop': '-18px',
                            # 'marginLeft': '20px',
                            'height': '25px',
                            'width': '100%',
                            "paddingBottom": "7px",
                            "paddingLeft": "20px",
                            "paddingRight": "20px",
                            'border':'1.5px white solid',
                            'backgroundColor': lightbrown,
                            "color": darkergrey,
                            "fontWeight": "bold",
                            "fontSize": "11px"}
carousel_image_style = {
                        #     'height': '50%',
                        # 'width': '100%',
                        'objectFit': 'contain',
                        }

white = '#ffffff'
bluewhite = "#dcebfc"
solidblue = "#27548f"

tab_selected_style2 = {'background': solidblue, 
                        'width': '15%',
                        'fontSize': '14px',
                        'color': white,
                        'fontWeight': 500,
                        'alignItems': 'center',
                        'justifyContent': 'center',
                        'padding':'7px',
                        'borderTop': '2px solid #27548f',
                        'borderLeft': '1px solid #27548f',
                        'borderBottom': '5px solid #f2810a'} # #d6d6d6


tab_selected_style3 = {'background': solidblue, 
                        'width': '30%',
                         'fontSize': '14px',
                         'color': white,
                         'fontWeight': 500,
                         'alignItems': 'center',
                         'justifyContent': 'center',
                         'padding':'7px',
                         'borderTop': '2px solid #27548f',
                         'borderLeft': '1px solid #27548f',
                         'borderBottom': '5px solid #f2810a'} # #d6d6d6


# -----------------------------------------------
# HTML components.
# -----------------------------------------------

# def banner_card():

#     return html.Div(
#             id="banner",
#             className="banner",
#             children=[html.Img(src=app.get_asset_url("pnnl_abbr_logo.png"))],
#         ),

def description_card():

    # -----------------------------------------------------------------------
    # A Div containing dashboard title & descriptions.
    # -----------------------------------------------------------------------

    return html.Div(
        id="description-card",
        children=[
            html.Img(id="logo", src=app.get_asset_url('logo3.png')),
            html.Hr(id="hr-break1"),
        ],
    )

def generate_control_card():

    # -----------------------------------------------------------------------
    # A Div containing controls for graphs. Controls include the following:
    #
    #   Scenario buttons, sliders, and dropdowns.
    # -----------------------------------------------------------------------

    return html.Div(
        id="control-card",
        children=[

            html.Div(id='scenario1-div',
                     children=[
                        html.Button('Run At Commercial Scale', 
                        id='btn-nclicks-1', n_clicks=0),
                        html.P(scenario1_text, id="scenario1_text"), 
            ]),
            html.Div(id='scenario2-div',
                     children=[
                        html.Button('Optimize Power Output', 
                        id='btn-nclicks-2', n_clicks=0),
                        html.P(scenario2_text, id="scenario2_text"), 
            ]),
            html.Div(id='scenario3-div',
                     children=[
                        html.Button('Optimize Economic Competitiveness', 
                        id='btn-nclicks-3', n_clicks=0),
                        html.P(scenario3_text, id="scenario3_text"),
            ]),            
            html.Hr(id="hr-break2"), 
            dropdown_card(),
            html.Br(),
            html.Br(),
            slider_card(),
            # disclaimer
            html.P(disclaimer_text, id='disclaimer-text'),
        ], 
    )



def graph_guidance_card(btnID, cardID, dropdown_children):

    # -----------------------------------------------------------------------
    # A Div containing graph guidance dropdowns for extra user context.
    # -----------------------------------------------------------------------

    return html.Div(
        [
            dbc.Button(
                "ⓘ Graph Guidance",
                id=btnID,
                color="primary",
                n_clicks=0,
                style=dropdown_guidance_style
            ),
            dbc.Collapse(
                dbc.Card(
                    children=[
                        dbc.CardBody(
                            children=[
                                dropdown_children
                            ]
                        )]
                    ),
                id=cardID,
                is_open=False,
            ),
        ]
    )



def generate_tabs():

    about_tab = dcc.Tab(label='About',
                        id="about-tab",
                         value='about-tab',
                         selected_style=tab_selected_style2,
                         children=[html.Hr(),
                                   html.Div(
                                        id="demo",
                                        style={'display': 'flex'},
                                        children=[
                                                  html.Div(id='intro2', 
                                                            children=[html.P("About Our Research", id='ab-title1'),
                                                                      html.P(note, id='ab-note'), 
                                                                      html.P(note2, style={"padding": "15px"}), # *****
                                                                      html.P("Navigating the Results", id='ab-title2'),
                                                                      html.P(note3, id='ab-note3'), 
                                                                      html.P("Resources", id='ab-title3'),
                                                                      html.P(note4, id='ab-note4'),
                                                    ]),
                                                    html.Div(id='image-container', 
                                                            children=[
                                                                html.Img(id="cluster-img", src=app.get_asset_url('CLGWG3.png'),
                                                                         # style=collab_image_style
                                                                                ),
                                                                      dbc.Carousel(id="carousel-ride",
                                                                            items=[
                                                                                {"key": "1", 
                                                                                 "src": app.get_asset_url('CLGS-1.3.png'), 
                                                                                 "img_style": carousel_image_style
                                                                                    },
                                                                                {"key": "2", 
                                                                                 "src": app.get_asset_url('CLGS-3.png'),
                                                                                 "img_style": carousel_image_style
                                                                                },
                                                                            ],
                                                                            variant="dark",
                                                                            ride="carousel",
                                                                            interval=3000,
                                                                            controls=True,
                                                                            indicators=False),
                                                                      ]
                                                            )
                                                  ],
                                        ),
                    ])


    energy_time_tab = dcc.Tab(label='Subsurface Results',
                         id="subsurface-tab",
                         value='energy-time-tab',
                         selected_style=tab_selected_style3,
                         children=[
                             graph_guidance_card(btnID="collapse-button2", cardID="collapse2", dropdown_children=html.P(dropdown_text1)),
                             html.Div(id="error_block_div1"), 
                             html.Br(),
                             html.Div(id="graphics-container",
                                        children=[
                                             dcc.Graph(id="geothermal_time_plots",
                                                      config=plotly_config),
                                            dbc.RadioItems(
                                                        options=[
                                                            {"label": "Auto Scale", "value": 1},
                                                            {"label": "Full Scale", "value": 2},
                                                        ],
                                                        value=1,
                                                        id="radio-graphic-control3",
                                                    ),
                                        ]

                                    )

                    ])


    energy_tab = dcc.Tab(label='Subsurface Contours',
                         id="contour-tab",
                         value='energy-tab',
                         selected_style=tab_selected_style3,
                         children=[
                             # html.Hr(),
                             html.Div(id="dropdown-card5",
                                      children=[
                                            graph_guidance_card(btnID="collapse-button", cardID="collapse", dropdown_children=html.P(dropdown_text2)),
                                            html.Div(id="error_block_div2"), 
                                            html.P("Select Parameter", id='select-text'),
                                            dcc.Dropdown(
                                                id="param-select",
                                                options=[{"label": i, "value": i} for i in param_list],
                                                value="Horizontal Extent (m)",
                                                clearable=False,
                                                searchable=False,
                                                style={'width': '60%',  # ***
                                                # 'marginLeft': '10px'
                                                }
                                            ),
                                        ]
                                    ), 
                            dcc.Graph(id="geothermal_plots",
                                      config=plotly_config)

                    ])

    economics_time_tab = dcc.Tab(label='Economic Results',
                            id="econ-tab",
                            value='economics-time-tab',
                            selected_style=tab_selected_style3,
                             children=[
                             # html.Hr(),
                             graph_guidance_card(btnID="collapse-button4", cardID="collapse4", 
                                                    dropdown_children=
                                                        html.Div(
                                                            children=[
                                                                html.P(dropdown_text1),
                                                                html.P(dropdown_econ_text1),
                                                                html.Img(id="formulas-img", src=app.get_asset_url('lcoe-lcoh-formulas.png')),
                                                                dcc.Markdown(dropdown_econ_markdown_text1, mathjax=True),
                                                                html.P(dropdown_econ_text2)
                                                                ]
                                                            )
                                                        ),
                             html.Div(id="warning_block_div3"),
                             html.Div(id="error_block_div3"), 
                             html.Br(),
                             # dcc.Graph(id="econ_plots",
                             #          config=plotly_config)
                             html.Div(id="graphics-container-econ", # was a duplicate id with graphics-container
                                        children=[
                                            dcc.Graph(id="econ_plots",
                                                        config=plotly_config),
                                            dbc.RadioItems(
                                                        options=[
                                                            {"label": "Auto Scale", "value": 1},
                                                            {"label": "Full Scale", "value": 2},
                                                        ],
                                                        value=1,
                                                        id="radio-graphic-control4",
                                                    ),
                                        ]

                                    )

                    ])

    summary_tab = dcc.Tab(label='Summary',
                            id="summary-tab",
                            value='summary-tab',
                            selected_style=tab_selected_style2,
                             children=[
                             html.Hr(),
                             html.Button("Download Results", id="btn_xlsx"),
                             dcc.Download(id="download-dataframe-xlsx"),
                             html.Br(),
                             html.Br(),
                             dcc.Graph(id="table",
                                      config=plotly_config)

                    ])


    tabs = dcc.Tabs(id="tabs", 
                    value='about-tab', 
                    children=[about_tab,
                              energy_time_tab,
                              energy_tab,
                              economics_time_tab,
                              summary_tab,
                              ])

    return tabs



# -----------------------------------------------------------------------------
# Define dash app layout here.
# -----------------------------------------------------------------------------

app.layout = html.Div(
    id="app-container",
    children=[
        dcc.Store(id='econ-memory'),
        dcc.Store(id='econ-results'),
        dcc.Store(id='econ-errors'),
        dcc.Store(id='thermal-memory'),
        dcc.Store(id='thermal-results-mass'),
        dcc.Store(id='thermal-results-time'),
        dcc.Store(id='thermal-results-errors'),
        dcc.Store(id='thermal-contours-errors'),
        dcc.Store(id='summary-memory'),

        # banner_card(),

        # Left column
        html.Div(
            id="left-column",
            className="four columns",
            children=[description_card(), 
                      generate_control_card()
                      ]
        ),
        # Right column
        html.Div(
            id="right-column",
            className="eight columns",
            children=[
                html.Div(
                    id="geothermal_card",
                    children=[generate_tabs()
                    ],
                ),
            ],
        ),
    ],
)


# -----------------------------------------------------------------------------
# Define dash app callbacks begin here.
# -----------------------------------------------------------------------------

@app.callback(
    Output(component_id="download-dataframe-xlsx", component_property="data"),
    [Input(component_id="btn_xlsx", component_property="n_clicks"),
     Input(component_id='summary-memory', component_property='data'),
     Input(component_id='thermal-results-mass', component_property='data'),
     Input(component_id='thermal-results-time', component_property='data'),
     Input(component_id='econ-results', component_property='data')],
    prevent_initial_call=True,
)

def func(n_clicks, df_tbl, df_mass_flow_rate, df_time, df_econ):

    if "btn_xlsx" == ctx.triggered_id:
        write_excelsheet(df_summary=df_tbl, df_subsurf_res_mass=df_mass_flow_rate, df_subsurf_res_time=df_time, df_econ=df_econ)
        return dcc.send_file('geoCLUSTER_results.xlsx') #dcc.Download('geoCLUSTER_results.xlsx') #dcc.send_file('geoCLUSTER_results.xlsx')
    else:
        raise PreventUpdate


@app.callback(
    Output(component_id="collapse", component_property="is_open"),
    [Input(component_id="collapse-button", component_property="n_clicks")],
    [State(component_id="collapse", component_property="is_open")],
)
def toggle_collapse(n, is_open):

    # ----------------------------------------------
    # Subsurface contours graph guidance.
    # ----------------------------------------------

    if n:
        return not is_open
    return is_open

@app.callback(
    Output(component_id="collapse2", component_property="is_open"),
    [Input(component_id="collapse-button2", component_property="n_clicks")],
    [State(component_id="collapse2", component_property="is_open")],
)
def toggle_collapse(n, is_open):

    # ----------------------------------------------
    # Subsurface results graph guidance.
    # ----------------------------------------------

    if n:
        return not is_open
    return is_open


@app.callback(
    Output(component_id="collapse3", component_property="is_open"),
    [Input(component_id="collapse-button3", component_property="n_clicks")],
    [State(component_id="collapse3", component_property="is_open")],
)
def toggle_collapse(n, is_open):

    # ----------------------------------------------
    # Open to finetune system values.
    # ----------------------------------------------

    if n:
        return not is_open
    return is_open


@app.callback(
    Output(component_id="collapse4", component_property="is_open"),
    [Input(component_id="collapse-button4", component_property="n_clicks")],
    [State(component_id="collapse4", component_property="is_open")],
)
def toggle_collapse(n, is_open):

    # ----------------------------------------------
    # Economic results graph guidance.
    # ----------------------------------------------

    if n:
        return not is_open
    return is_open


@app.callback(
   [Output(component_id='fluid-select', component_property='options'),
    Output(component_id='fluid-select', component_property='value'),
    Output(component_id='interpolation-select', component_property='options'),
    Output(component_id='interpolation-select', component_property='value')
   ],
   [Input(component_id="tabs", component_property="value")])

def change_dropdown(at):

    if at == "energy-time-tab":
        fluid_list = ["All", "H2O", "sCO2"]
        interpol_list = ["True"]
        return [{"label": i, "value": i} for i in fluid_list], fluid_list[0], [{"label": i, "value": i} for i in interpol_list], interpol_list[0]

    elif at == "about-tab":
        fluid_list = ["All", "H2O", "sCO2"]
        interpol_list = ["True"]
        return [{"label": i, "value": i} for i in fluid_list], fluid_list[0], [{"label": i, "value": i} for i in interpol_list], interpol_list[0]
   
    elif at == "energy-tab":
        fluid_list = ["H2O", "sCO2"]
        interpol_list = ["True"]
        return [{"label": i, "value": i} for i in fluid_list], fluid_list[0], [{"label": i, "value": i} for i in interpol_list], interpol_list[0]

    elif at == "economics-time-tab":
        fluid_list = ["All", "H2O", "sCO2"]
        interpol_list = ["True"]
        return [{"label": i, "value": i} for i in fluid_list], fluid_list[0], [{"label": i, "value": i} for i in interpol_list], interpol_list[0]

    elif at == "summary-tab":
        fluid_list = ["All", "H2O", "sCO2"]
        interpol_list = ["True"]
        return [{"label": i, "value": i} for i in fluid_list], fluid_list[0], [{"label": i, "value": i} for i in interpol_list], interpol_list[0]
        # raise PreventUpdate


@app.callback(
    [Output(component_id='mdot-select', component_property='value'),
     Output(component_id='L2-select', component_property='value'),
     Output(component_id='L1-select', component_property='value'),
     Output(component_id='grad-select', component_property='value'),
     Output(component_id='diameter-select', component_property='value'),
     Output(component_id='Tinj-select', component_property='value'),
     Output(component_id='k-select', component_property='value'),

     Output(component_id='drillcost-select', component_property='value'),
     Output(component_id='discount-rate-select', component_property='value'),
     Output(component_id='lifetime-select', component_property='value'),
     Output(component_id='kwt-select', component_property='value'),
     Output(component_id='kwe-select', component_property='value'),
     Output(component_id='precool-select', component_property='value'),
     Output(component_id='turb-pout-select', component_property='value'),
   ],
    [Input(component_id='btn-nclicks-1', component_property='n_clicks'),
     Input(component_id='btn-nclicks-2', component_property='n_clicks'),
     Input(component_id='btn-nclicks-3', component_property='n_clicks'),
     Input(component_id="tabs", component_property="value"),
     Input(component_id='case-select', component_property='value'),
     Input(component_id='fluid-select', component_property='value'),
     Input(component_id='end-use-select', component_property='value')
     ]
)

def update_slider_with_btn(btn1, btn2, btn3, at, case, fluid, end_use):

    # ----------------------------------------------------------------------------------------------
    # Defines scenario button values when clicked.  
    #
    # Commercial Scale:
    # For example, consider a u-shaped configuration installed in a geothermal reservoir, with a 
    # horizontal extent of 10 km, vertical depth of 2.5 km, geothermal gradient of 0.065 ˚C/m, borehole 
    # diameter of 0.3 m, inlet temperature of 40˚C, and rock thermal conductivity of 3.5 W/m K, 
    # parameters within the range of anticipated values for commercial scale geothermal applications
    #
    #
    # ----------------------------------------------------------------------------------------------

    # output = ('utube', 24, 10000, 3500, 0.050, 0.35, 30, 3, 1000, 7.0, 40, 100, 3000, 13, 80) # return to default

    if "btn-nclicks-1" == ctx.triggered_id:
        output = (24, 10000, 2500, 0.050, 0.30, 40, 3.5,  1000, 7.0, 40, 100, 3000, 13, 80)
        return output
    
    elif "btn-nclicks-2" == ctx.triggered_id: 
        if case == 'coaxial':
            output = (77, 20000, 5000, 0.070, 0.44, 30, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            return output
        if case == "utube":
            output = (100, 20000, 5000, 0.070, 0.44, 30, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            return output

    elif "btn-nclicks-3" == ctx.triggered_id:

        if case == 'coaxial' and fluid == "H2O" or fluid == "All":

            if end_use == "Electricity":
                output = (39.2, 20000, 5000, 0.070, 0.444, 60, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            if end_use == "Heating" or end_use == "All":
                output = (73.4, 13000, 5000, 0.070, 0.444, 30, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            return output

        if case == 'utube' and fluid == "H2O" or fluid == "All":

            if end_use == "Electricity":
                output = (43, 20000, 5000, 0.070, 0.44, 60, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            if end_use == "Heating" or end_use == "All":
                output = (100, 20000, 5000, 0.070, 0.44, 30, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            return output

        if case == 'coaxial' and fluid == "sCO2":
            
            if end_use == "Electricity":
                output = (69.6, 13000, 5000, 0.070, 0.44, 60, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            if end_use == "Heating" or end_use == "All":
                output = (69.6, 6000, 5000, 0.070, 0.44, 45, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            return output
            
        if case == 'utube' and fluid == "sCO2":
            
            if end_use == "Electricity":
                output = (100, 20000, 5000, 0.070, 0.44, 60, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            if end_use == "Heating" or end_use == "All":
                output = (100, 11000, 5000, 0.070, 0.44, 45, 4.5,  1000, 7.0, 40, 100, 3000, 13, 80)
            return output

    else:
        raise PreventUpdate



@app.callback(
   [Output(component_id='mdot-select-div', component_property='style'),
    Output(component_id='L2-select-div', component_property='style'),
    Output(component_id='L1-select-div', component_property='style'),
    Output(component_id='grad-select-div', component_property='style'),
    Output(component_id='diameter-select-div', component_property='style'),
    Output(component_id='Tinj-select-div', component_property='style'),
    Output(component_id='k-select-div', component_property='style'),

    Output(component_id='drillcost-div', component_property='style'),
    Output(component_id='discount-rate-div', component_property='style'),
    Output(component_id='lifetime-div', component_property='style'),
    Output(component_id='kwt-div', component_property='style'),
    Output(component_id='kwe-div', component_property='style'),
    Output(component_id='precool-div', component_property='style'),
    Output(component_id='turb-pout-div', component_property='style'),

   ],
   [Input(component_id='param-select', component_property='value'),
    Input(component_id="tabs", component_property="value"),
    Input(component_id="fluid-select", component_property="value"),
    Input(component_id="end-use-select", component_property="value")
    ])

def show_hide_element(visibility_state, at, fluid, end_use):

    # ----------------------------------------------------------------------------------------------
    # Reveals or hides sliders depending on which tab selected and which dropdowns.
    # ----------------------------------------------------------------------------------------------

    if at == "about-tab":
        if fluid == "H2O":
            return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'},\
                    {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'none'}
        else:
            return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'},\
                    {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}
    
    elif at == "energy-time-tab":
        return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}
    
    elif at == "energy-tab":

        if visibility_state == param_list[0]:
            return {'display': 'none'}, {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                    {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}
        if visibility_state == param_list[1]:
            return {'display': 'none'}, {'display': 'block'}, {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                    {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}
        if visibility_state == param_list[2]:
            return {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                    {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}
        if visibility_state == param_list[3]:
            return {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, \
                    {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}
        if visibility_state == param_list[4]:
            return {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'block'}, \
                    {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}
        if visibility_state == param_list[5]:
            return {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, \
                    {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}
    
    elif at == "economics-time-tab":
        if fluid == "H2O":
            if end_use == "All":
                return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                        {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'none'}
            if end_use == "Heating":
                return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                        {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'none'}, {'display': 'none'}
            if end_use == "Electricity":
                return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                        {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'block'}, {'display': 'none'}, {'display': 'none'}

        else:
            if end_use == "All":
                return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                        {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}
            if end_use == "Heating":
                return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                        {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'block'}, {'display': 'block'}
            if end_use == "Electricity":
                return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                        {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}

    elif at == "summary-tab":
        if fluid == "H2O":
            return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                    {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'none'}, {'display': 'none'}
        else:
            return {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, \
                    {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}, {'display': 'block'}
    else:
        raise PreventUpdate



# -----------------------------------------------------------------------------
# Define dash app plotting callbacks.
# -----------------------------------------------------------------------------

@app.callback(
    [Output(component_id="geothermal_time_plots", component_property="figure"),
     Output(component_id='thermal-memory', component_property='data'),
     Output(component_id='thermal-results-mass', component_property='data'),
     Output(component_id='thermal-results-time', component_property='data'),
     Output(component_id='thermal-results-errors', component_property='data')
     ],
    [Input(component_id="interpolation-select", component_property="value"),
     Input(component_id="fluid-select", component_property="value"),
     Input(component_id="case-select", component_property="value"),
     Input(component_id="mdot-select", component_property="value"),
     Input(component_id="L2-select", component_property="value"),
     Input(component_id="L1-select", component_property="value"),
     Input(component_id="grad-select", component_property="value"),
     Input(component_id="diameter-select", component_property="value"),
     Input(component_id="Tinj-select", component_property="value"),
     Input(component_id="k-select", component_property="value"),

     Input(component_id='radio-graphic-control3', component_property='value'),
    ],
)

def update_subsurface_results_plots(interp_time, fluid, case, mdot, L2, L1, grad, D, Tinj, k, scale):

    # -----------------------------------------------------------------------------
    # Creates and displays Plotly subplots of the subsurface results.
    # -----------------------------------------------------------------------------

    subplots, forty_yr_TPmeans_dict, df_mass_flow_rate, df_time, err_subres_dict = generate_subsurface_lineplots(
        interp_time, fluid, case, mdot, L2, L1, grad, D, Tinj, k, scale
    )

    return subplots, forty_yr_TPmeans_dict, df_mass_flow_rate, df_time, err_subres_dict



@app.callback(
    [Output(component_id="geothermal_plots", component_property="figure"),
     Output(component_id="thermal-contours-errors", component_property="data"),
    ],
    [Input(component_id="interpolation-select", component_property="value"),
     Input(component_id="fluid-select", component_property="value"),
     Input(component_id="case-select", component_property="value"),
     Input(component_id="param-select", component_property="value"),
     Input(component_id="mdot-select", component_property="value"),
     Input(component_id="L2-select", component_property="value"),
     Input(component_id="L1-select", component_property="value"),
     Input(component_id="grad-select", component_property="value"),
     Input(component_id="diameter-select", component_property="value"),
     Input(component_id="Tinj-select", component_property="value"),
     Input(component_id="k-select", component_property="value"),
    ],
)

def update_subsurface_contours_plots(interp_time, fluid, case, param, mdot, L2, L1, grad, D, Tinj, k):

    # -----------------------------------------------------------------------------
    # Creates and displays Plotly subplots of the subsurface contours.
    # -----------------------------------------------------------------------------

    subplots, err_subcontour_dict = generate_subsurface_contours(
        interp_time, fluid, case, param, mdot, L2, L1, grad, D, Tinj, k
    )

    return subplots, err_subcontour_dict


@app.callback(
    [Output(component_id="econ_plots", component_property="figure"),
     Output(component_id='econ-memory', component_property='data'),
     Output(component_id='econ-results', component_property='data'),
     Output(component_id='econ-errors', component_property='data')
     ],
    [Input(component_id="interpolation-select", component_property="value"),
     Input(component_id="fluid-select", component_property="value"),
     Input(component_id="case-select", component_property="value"),
     Input(component_id="end-use-select", component_property="value"),

     Input(component_id="mdot-select", component_property="value"),
     Input(component_id="L2-select", component_property="value"),
     Input(component_id="L1-select", component_property="value"),
     Input(component_id="grad-select", component_property="value"),
     Input(component_id="diameter-select", component_property="value"),
     Input(component_id="Tinj-select", component_property="value"),
     Input(component_id="k-select", component_property="value"),

     Input(component_id="drillcost-select", component_property="value"),
     Input(component_id="discount-rate-select", component_property="value"),
     Input(component_id="lifetime-select", component_property="value"),
     Input(component_id="kwt-select", component_property="value"),
     Input(component_id="kwe-select", component_property="value"),
     Input(component_id="precool-select", component_property="value"),
     Input(component_id="turb-pout-select", component_property="value"),

     Input(component_id='radio-graphic-control4', component_property='value'),
    ],
)

def update_econ_plots(interp_time, fluid, case, end_use,
                      mdot, L2, L1, grad, D, Tinj, k,
                      Drilling_cost_per_m, Discount_rate, Lifetime, 
                      Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure,
                      scale):

    # -----------------------------------------------------------------------------
    # Creates and displays Plotly subplots of the economic results.
    # -----------------------------------------------------------------------------


    economics_fig, econ_data_dict, econ_values_dict, err_econ_dict = generate_econ_lineplots(
        interp_time, case, end_use, fluid, 
        mdot, L2, L1, grad, D, Tinj, k,
        Drilling_cost_per_m, Discount_rate, Lifetime, 
        Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure,
        scale
    )

    return economics_fig, econ_data_dict, econ_values_dict, err_econ_dict


@app.callback(
    # Output(component_id="table", component_property="figure"),
     [Output(component_id="table", component_property="figure"),
      Output(component_id='summary-memory', component_property='data')
     ],
    [Input(component_id="interpolation-select", component_property="value"),
     Input(component_id="fluid-select", component_property="value"),
     Input(component_id="case-select", component_property="value"),
     # Input("end-use-select", "value"),

     Input(component_id="mdot-select", component_property="value"),
     Input(component_id="L2-select", component_property="value"),
     Input(component_id="L1-select", component_property="value"),
     Input(component_id="grad-select", component_property="value"),
     Input(component_id="diameter-select", component_property="value"),
     Input(component_id="Tinj-select", component_property="value"),
     Input(component_id="k-select", component_property="value"),

     Input(component_id="drillcost-select", component_property="value"),
     Input(component_id="discount-rate-select", component_property="value"),
     Input(component_id="lifetime-select", component_property="value"),
     Input(component_id="kwt-select", component_property="value"),
     Input(component_id="kwe-select", component_property="value"),
     Input(component_id="precool-select", component_property="value"),
     Input(component_id="turb-pout-select", component_property="value"),
     Input(component_id='econ-memory', component_property='data'),
     Input(component_id='thermal-memory', component_property='data'),
    ],
)

def update_table(interp_time, fluid, case, mdot, L2, L1, grad, D, Tinj, k,
                 Drilling_cost_per_m, Discount_rate, Lifetime, 
                 Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure,
                 econ_dict, thermal_dict):

    tbl, summary_dict = generate_summary_table(
                mdot, L2, L1, grad, D, Tinj, k, Drilling_cost_per_m, Discount_rate, Lifetime, 
                Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure, 
                interp_time, case, fluid,
                thermal_dict, econ_dict
    )

    return tbl, summary_dict


@app.callback(
    [Output(component_id='error_block_div1', component_property='children'),
     Output(component_id='error_block_div2', component_property='children'),
     Output(component_id='error_block_div3', component_property='children'),
     ],
    [Input(component_id='thermal-results-errors', component_property='data'),
     Input(component_id='thermal-contours-errors', component_property='data'),
     Input(component_id='econ-errors', component_property='data'),
    ]
)

def update_error_divs(err_sub_dict, err_contour_dict, err_econ_dict):
    
    # print(err_sub_dict)
    # print(err_contour_dict)
    # print(err_econ_dict)
    # print('\n')

    err_div1 = html.Div(#id="error_block_div1",
                        style={'display': 'none'})

    err_div2 = html.Div(#id="error_block_div2",
                        style={'display': 'none'})

    err_div3 = html.Div(#id="error_block_div3",
                        style={'display': 'none'})

    error_style = {'display': 'block',
                    'width': '100%',
                    "paddingTop": "8px",
                    "paddingLeft": "20px",
                    "paddingRight": "20px",
                    "paddingBottom": "5px",
                    'backgroundColor': lightbrown,
                    'color': darkergrey,
                    }
    if err_sub_dict != {}:


        error_message = next(iter(err_sub_dict.values()))

        if "No outputs" in error_message:
            error_message = "No outputs were able to be calculated because there are not enough data at these limits. Consider changing parameter value(s)."

        err_div1 = html.Div(#id="error_block_div1",
                            style=error_style,
                            children=[
                                html.Img(id="error-img1", src=app.get_asset_url('error.png')), 
                                dcc.Markdown("**Did not plot visual(s).**", style={'display': 'inline-block'}),
                                html.P(error_message),
                                ]
                        )


    if err_contour_dict != {}:

        error_message = next(iter(err_contour_dict.values()))

        err_div2 = html.Div(#id="error_block_div2",
                            style=error_style,
                            children=[
                                html.Img(id="error-img2", src=app.get_asset_url('error.png')), 
                                dcc.Markdown("**Did not plot visual(s).**", style={'display': 'inline-block'}),
                                html.P(error_message),
                                ]
                        )

    if err_econ_dict != {}:

        error_message = next(iter(err_econ_dict.values()))

        if "object has no attribute" in error_message:
            error_message = "No outputs were able to be calculated because there are not enough data at these limits. Consider changing parameter value(s)."

        err_div3 = html.Div(#id="error_block_div3",
                            style=error_style,
                            children=[
                                html.Img(id="error-img3", src=app.get_asset_url('error.png')),
                                dcc.Markdown("**Did not plot visual(s).**", style={'display': 'inline-block'}),
                                html.P(error_message),
                                ]
                        )

    return err_div1, err_div2, err_div3


@app.callback(
    Output(component_id='warning_block_div3', component_property='children'),
    Input(component_id='econ-memory', component_property='data'),
)

def update_error_divs(levelized_cost_dict):
    
    warning_div3 = html.Div(style={'display': 'none'})

    error_style = {'display': 'block',
                    'width': '100%',
                    "paddingTop": "8px",
                    "paddingLeft": "20px",
                    "paddingRight": "20px",
                    "paddingBottom": "5px",
                    'backgroundColor': lightbrown,
                    'color': darkergrey,
                    }

    if levelized_cost_dict['LCOE sCO2'] == "9999.00" or levelized_cost_dict['LCOE H2O'] == "9999.00":
        
        warning_div3 = html.Div(#id="error_block_div3",
                            style=error_style,
                            children=[
                                html.Img(id="warning-img", src=app.get_asset_url('warning.png')), 
                                dcc.Markdown("**LCOE is too high.**", style={'display': 'inline-block'}),
                                html.P("Outlet Temperature (°C) may be too low or the system is losing heat (negative kWe)."),
                                ]
                        )


    return warning_div3



# -----------------------------------------------------------------------------
# App run here. Define configurations, proxies, etc.
# -----------------------------------------------------------------------------

server = app.server 
# from app import server as application # in the wsgi.py file -- this targets the Flask server of Dash app

if __name__ == '__main__':
    app.run_server(port=8060, debug=True) 
    # EXAMPLE: *************
    # app.run_server(port=8050, proxy="http://127.0.0.1:8059::https://<site>/<page_name>")




