#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# standard system libraries
import sys

# data manipulation libraries
import numpy as np

# plotting libraries
import plotly.graph_objects as go
import plotly.express as px
from plotly.subplots import make_subplots

# sourced scripts
from reader import initialize_data, data_dict
from econ_support import create_teaobject
from plots_support import * 

# -----------------------
# Read in data.
# -----------------------
u_sCO2, u_H2O, c_sCO2, c_H2O = initialize_data()
param_dict = data_dict(u_sCO2, u_H2O, c_sCO2, c_H2O)

# -----------------------
# Global properties.
# -----------------------
to_kelvin_factor = 273.15

labels_cat = ['Working Fluid2', 'Working Fluid1'] # legend visibility and syncing 
labels = ['H2O','sCO2'] # legend visibility and syncing 
lw = 1.4 # line width

# --------------------------------------------------------------------------------------------------------------------
# THERMAL PERFORMANCE PLOTS
# --------------------------------------------------------------------------------------------------------------------

def param_nearest_init(arg_mdot, arg_L2, arg_L1, arg_grad, arg_D, arg_Tinj, arg_k):

    arg_mdot_v, arg_mdot_i = find_nearest(u_sCO2.mdot, arg_mdot) 
    arg_L2_v, arg_L2_i = find_nearest(u_sCO2.L2, arg_L2)
    arg_L1_v, arg_L1_i = find_nearest(u_sCO2.L1, arg_L1)
    arg_grad_v, arg_grad_i = find_nearest(u_sCO2.grad, arg_grad)
    arg_D_v, arg_D_i = find_nearest(u_sCO2.D, arg_D)
    arg_Tinj_v, arg_Tinj_i = find_nearest(u_sCO2.Tinj, arg_Tinj + to_kelvin_factor) # to kelvin
    arg_k_v, arg_k_i = find_nearest(u_sCO2.k, arg_k)

    return arg_mdot_v, arg_mdot_i, arg_L2_v, arg_L2_i, arg_L1_v, arg_L1_i, arg_grad_v, arg_grad_i, arg_D_v, arg_D_i, \
                arg_Tinj_v, arg_Tinj_i, arg_k_v, arg_k_i


def get_kWe_kWt_over_mass_or_time(case, fluid, point, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i):

    # --------------------------------------------------------------------------------------------------------------
    # For both H2O and sCO2, calculate:
    #       - kWe and kWt averages over mass flow rate & 
    #       - instantaneous kWe and kWt overtime 
    #
    # Outputs can be used for when interp_time == True, the following exists:
    #           sCO2_kWe_avg, sCO2_kWt_avg, H2O_kWe_avg, H2O_kWt_avg
    # Then, the following also get overriden when interp_time == True:
    #           sCO2_kWe, sCO2_kWt, H2O_kWe, H2O_kWt
    # 
    # When interp_time == False then all the output will exist and be plotted.
    # --------------------------------------------------------------------------------------------------------------

    # need initialization here
    sCO2_kWe_avg = None
    sCO2_kWt_avg = None 
    H2O_kWe_avg = None
    H2O_kWt_avg = None
    error_messages_d = {}

    if case == "utube":

        try:
            sCO2_kWe, sCO2_kWt = u_sCO2.interp_kW(point)
            H2O_kWe, H2O_kWt = u_H2O.interp_kW(point)

        except ValueError as e:
            sCO2_kWe, sCO2_kWt, H2O_kWe, H2O_kWt = blank_data_kW()
            error_message = parse_error_message(e=e, e_name='Err SubRes1')
            error_messages_d['Err SubRes1'] = error_message

        if fluid == "H2O" or fluid == "All":
            H2O_kWe_avg = u_H2O.kWe_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i]
            H2O_kWt_avg = u_H2O.kWt_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i]

        if fluid == "sCO2" or fluid == "All":
            sCO2_kWe_avg = u_sCO2.kWe_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i]
            sCO2_kWt_avg = u_sCO2.kWt_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i]

    if case == "coaxial":

        try:
            sCO2_kWe, sCO2_kWt = c_sCO2.interp_kW(point)
            H2O_kWe, H2O_kWt = c_H2O.interp_kW(point)
        
        except ValueError as e:
            sCO2_kWe, sCO2_kWt, H2O_kWe, H2O_kWt = blank_data_kW()
            error_message = parse_error_message(e=e, e_name='Err SubRes2')
            error_messages_d['Err SubRes2'] = error_message

        if fluid == "H2O" or fluid == "All":
            H2O_kWe_avg = c_H2O.kWe_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i]
            H2O_kWt_avg = c_H2O.kWt_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i]

        if fluid == "sCO2" or fluid == "All":
            sCO2_kWe_avg = c_sCO2.kWe_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i]
            sCO2_kWt_avg = c_sCO2.kWt_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i]

    return sCO2_kWe, sCO2_kWt, H2O_kWe, H2O_kWt, sCO2_kWe_avg, sCO2_kWt_avg, H2O_kWe_avg, H2O_kWt_avg, error_messages_d


def generate_subsurface_lineplots(interp_time, fluid, case, arg_mdot, arg_L2, arg_L1, arg_grad, arg_D, arg_Tinj, arg_k, scale):

    # -----------------------------------------------------------------------------------------------------------------
    # Creates Plotly with 5 subplots:
    #
    #       kWe vs. mass flow rate
    #       kWt vs. mass flow rate
    #       Temperature vs. Time (years)
    #       Pressure vs. Time (years)
    #       kWe vs. time
    #
    # -----------------------------------------------------------------------------------------------------------------

    mean_H2O_Tout = "-"
    mean_H2O_Pout = "-"
    mean_sCO2_Tout = "-"
    mean_sCO2_Pout = "-"

    error_messages_dict = {}

    time = u_sCO2.time
    m_dot = u_sCO2.mdot

    mass_flow_rates_dict = {"Mass Flow Rate (kg/s)": m_dot}
    time_dict = {"Time (year)": time}

    arg_mdot_v, arg_mdot_i, arg_L2_v, arg_L2_i, arg_L1_v, arg_L1_i, arg_grad_v, arg_grad_i, arg_D_v, arg_D_i, \
                arg_Tinj_v, arg_Tinj_i, arg_k_v, arg_k_i = param_nearest_init(arg_mdot, arg_L2, arg_L1, arg_grad, arg_D, arg_Tinj, arg_k)

    point = (arg_mdot, arg_L2, arg_L1, arg_grad, arg_D, arg_Tinj + to_kelvin_factor, arg_k) # to kelvin

    # ** Average calculations
    sCO2_kWe, sCO2_kWt, H2O_kWe, H2O_kWt, sCO2_kWe_avg, sCO2_kWt_avg, H2O_kWe_avg, H2O_kWt_avg, error_messages_d = \
                            get_kWe_kWt_over_mass_or_time(case, fluid, point, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i)

    error_messages_dict.update(error_messages_d)

    # print(sCO2_kWe_avg, sCO2_kWt_avg, H2O_kWe_avg, H2O_kWt_avg)

    is_blank_data = False

    if interp_time == "False":

        # this doesn't impact the kWe over time

        if case == "utube":

            if fluid == "H2O" or fluid == "All":
                H2O_Tout = u_H2O.Tout[arg_mdot_i, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, :]
                H2O_Pout = u_H2O.Pout[arg_mdot_i, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, :]

            if fluid == "sCO2" or fluid == "All":
                sCO2_Tout = u_sCO2.Tout[arg_mdot_i, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, :]
                sCO2_Pout = u_sCO2.Pout[arg_mdot_i, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, :]

        if case == "coaxial":

            if fluid == "H2O" or fluid == "All":
                H2O_Tout = c_H2O.Tout[arg_mdot_i, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, :]
                H2O_Pout = c_H2O.Pout[arg_mdot_i, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, :]

            if fluid == "sCO2" or fluid == "All":
                sCO2_Tout = c_sCO2.Tout[arg_mdot_i, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, :]
                sCO2_Pout = c_sCO2.Pout[arg_mdot_i, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, :]

    if interp_time == "True":

        # this doesn't impact the kWe and kWt averages

        if case == "utube":

            try:
                sCO2_Tout, sCO2_Pout = u_sCO2.interp_outlet_states(point)
                H2O_Tout, H2O_Pout = u_H2O.interp_outlet_states(point)
                sCO2_kWe, sCO2_kWt = u_sCO2.interp_kW(point)
                H2O_kWe, H2O_kWt = u_H2O.interp_kW(point)

            except ValueError as e:
                sCO2_Tout, sCO2_Pout, H2O_Tout, H2O_Pout, sCO2_kWe, sCO2_kWt, H2O_kWe,H2O_kWt = blank_data()
                error_message = parse_error_message(e=e, e_name='Err SubRes3')
                error_messages_dict['Err SubRes3'] = error_message
                is_blank_data = True
                
        if case == "coaxial":

            try:
                sCO2_Tout, sCO2_Pout = c_sCO2.interp_outlet_states(point)
                H2O_Tout, H2O_Pout = c_H2O.interp_outlet_states(point)
                sCO2_kWe, sCO2_kWt = c_sCO2.interp_kW(point)
                H2O_kWe, H2O_kWt = c_H2O.interp_kW(point)

            except ValueError as e:
                sCO2_Tout, sCO2_Pout, H2O_Tout, H2O_Pout, sCO2_kWe, sCO2_kWt, H2O_kWe,H2O_kWt = blank_data()
                error_message = parse_error_message(e=e, e_name='Err SubRes4')
                error_messages_dict['Err SubRes4'] = error_message
                is_blank_data = True
                


    # Figure generation 

    fig = make_subplots(rows=2, cols=3,
                        specs=[[{}, {}, None],
                        [{}, {}, {}]],
                        horizontal_spacing = 0.11,
                        vertical_spacing = 0.21
                        )

    if fluid == "sCO2" or fluid == "All":

        # kWe_avg and kWt_avg
        fig.add_trace(go.Scatter(x=m_dot, y=sCO2_kWe_avg,
                      hovertemplate='<b>Mass Flow Rate (kg/s)</b>: %{x:.1f}<br><b>Average kWt</b>: %{y:.3f} ',
                      line = dict(color='#c26219', width=lw),
                      legendgroup=labels_cat[1], name=labels[1], showlegend=False),
                      # margin=dict(pad=20),
                      row=1, col=1)
        fig.add_trace(go.Scatter(x=m_dot, y=sCO2_kWt_avg,
                      hovertemplate='<b>Mass Flow Rate (kg/s)</b>: %{x:.1f}<br><b>Average kWt</b>: %{y:.3f} ',
                      line = dict(color='orange', width=lw),
                      legendgroup=labels_cat[1], name=labels[1], showlegend=False),
                      row=1, col=2)

        # kWe vs. time
        fig.add_trace(go.Scatter(x=time, y=sCO2_kWe,
                      hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>kWe</b>: %{y:.3f} ',
                      line = dict(color='black', width=lw), # #379dbf
                      legendgroup=labels_cat[1], name=labels[1], showlegend=True),
                      row=2, col=1)
        # temperature
        fig.add_trace(go.Scatter(x=time, y=sCO2_Tout - to_kelvin_factor, # to celsius
                      hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Outlet Temperature (˚C)</b>: %{y:.3f} ',
                      line = dict(color='royalblue', width=lw),
                      legendgroup=labels_cat[1], name=labels[1], showlegend=False),
                      row=2, col=2)
        # pressure
        fig.add_trace(go.Scatter(x=time, y=sCO2_Pout / 1000000,
                      hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Outlet Pressure (MPa)</b>: %{y:.3f} ',
                      line = dict(color='#16b8a2', width=lw),
                      legendgroup=labels_cat[1], name=labels[1], showlegend=False),
                      row=2, col=3)

        if is_blank_data:
            blank_canvas(fig=fig, row_n=2, col_n=1)
            blank_canvas(fig=fig, row_n=2, col_n=2)
            blank_canvas(fig=fig, row_n=2, col_n=3)

        mean_sCO2_Tout = round(np.mean(sCO2_Tout - to_kelvin_factor),2) # to celsius
        mean_sCO2_Pout = round(np.mean(sCO2_Pout / 1000000),2)

        if str(mean_sCO2_Tout) == "nan":
            mean_sCO2_Tout = "-"
        if str(mean_sCO2_Pout) == "nan":
            mean_sCO2_Pout = "-"

        mass_flow_rates_dict["sCO2 40-Year Average of Exergy (kWe)"] = sCO2_kWe_avg
        mass_flow_rates_dict["sCO2 40-Year Average of Available Thermal Output (kWt)"] = sCO2_kWt_avg

        time_dict["sCO2 Exergy (kWe)"] = sCO2_kWe
        time_dict["sCO2 Outlet Temperature (˚C)"] = sCO2_Tout - to_kelvin_factor # to celsius
        time_dict["sCO2 Outlet Pressure (MPa)"] = sCO2_Pout / 1000000

    if fluid == "H2O" or fluid == "All":

        # kWe_avg and kWt_avg
        fig.add_trace(go.Scatter(x=m_dot, y=H2O_kWe_avg,
                      hovertemplate='<b>Mass Flow Rate (kg/s)</b>: %{x:.1f}<br><b>Average kWe</b>: %{y:.3f} ',
                      line = dict(color='#c26219', width=lw, dash='dash'),
                      legendgroup=labels_cat[0], name=labels[0], showlegend=False),
                      # margin=dict(pad=20),
                      row=1, col=1)

        fig.add_trace(go.Scatter(x=m_dot, y=H2O_kWt_avg,
                      hovertemplate='<b>Mass Flow Rate (kg/s)</b>: %{x:.1f}<br><b>Average kWt</b>: %{y:.3f} ',
                      line = dict(color='orange', width=lw, dash='dash'),
                      legendgroup=labels_cat[0], name=labels[0], showlegend=False),
                      row=1, col=2)

        # kWe vs. time
        fig.add_trace(go.Scatter(x=time, y=H2O_kWe, 
                      hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>kWe</b>: %{y:.3f} ',
                      line = dict(color='black', width=lw, dash='dash'), # #379dbf
                      legendgroup=labels_cat[0], name=labels[0], showlegend=True),
                      row=2, col=1)
        # temperature
        fig.add_trace(go.Scatter(x=time, y=H2O_Tout - to_kelvin_factor, # to celsius
                      hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Outlet Temperature (˚C)</b>: %{y:.3f} ',
                      line = dict(color='royalblue', width=lw, dash='dash'),
                      legendgroup=labels_cat[0], name=labels[0], showlegend=False),
                      row=2, col=2)
        # pressure
        fig.add_trace(go.Scatter(x=time, y=H2O_Pout / 1000000,
                      hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Outlet Pressure (MPa)</b>: %{y:.3f} ',
                      line = dict(color='#16b8a2', width=lw, dash='dash'),
                      legendgroup=labels_cat[0], name=labels[0], showlegend=False),
                      row=2, col=3)

        if is_blank_data:
            blank_canvas(fig=fig, row_n=2, col_n=1)
            blank_canvas(fig=fig, row_n=2, col_n=2)
            blank_canvas(fig=fig, row_n=2, col_n=3)

        mean_H2O_Tout = round(np.mean(H2O_Tout - to_kelvin_factor),2) # to celsius
        mean_H2O_Pout = round(np.mean(H2O_Pout / 1000000),2)

        if str(mean_H2O_Tout) == "nan":
            mean_H2O_Tout = "-"
        if str(mean_H2O_Pout) == "nan":
            mean_H2O_Pout = "-"

        mass_flow_rates_dict["H2O 40-Year Average of Exergy (kWe)"] = H2O_kWe_avg
        mass_flow_rates_dict["H2O 40-Year Average of Available Thermal Output (kWt)"] = H2O_kWt_avg

        time_dict["H2O Exergy (kWe)"] = H2O_kWe
        time_dict["H2O Outlet Temperature (˚C)"] = H2O_Tout - to_kelvin_factor # to celsius
        time_dict["H2O Outlet Pressure (MPa)"] = H2O_Pout / 1000000

    
    fig = update_layout_properties_subsurface_results(fig=fig, m_dot=m_dot, time=time, plot_scale=scale)

    forty_yr_means_dict = {'Mean H2O Tout': mean_H2O_Tout, 
                            'Mean H2O Pout': mean_H2O_Pout,
                            'Mean sCO2 Tout': mean_sCO2_Tout,
                            'Mean sCO2 Pout': mean_sCO2_Pout}

    return fig, forty_yr_means_dict, mass_flow_rates_dict, time_dict, error_messages_dict



def generate_subsurface_contours(interp_time, fluid, case, param, arg_mdot, arg_L2, arg_L1, arg_grad, arg_D, arg_Tinj, arg_k):

    # -----------------------------------------------------------------------------------------------------------------
    # Creates Plotly with 4 subplot contours:
    #
    #       <parameter selected> vs. mass flow rate vs. kWe
    #       <parameter selected> vs. mass flow rate vs. kWt
    #       <parameter selected> vs. mass flow rate vs. Tout
    #       <parameter selected> vs. mass flow rate vs. Pout
    #
    # -----------------------------------------------------------------------------------------------------------------

    error_messages_dict = {}

    # initializer to prevent errors
    if fluid == "All":
        fluid = "H2O"

    # contour elements
    param_y = param_dict[(case, fluid, param)]
    mdot_x = param_dict[(case, fluid, "mdot")]
    mdot_ij, param_ij = np.meshgrid(mdot_x, param_y, indexing='ij') # returns coordinate matrices from coordinate vectors

    arg_mdot_v, arg_mdot_i, arg_L2_v, arg_L2_i, arg_L1_v, arg_L1_i, arg_grad_v, arg_grad_i, arg_D_v, arg_D_i, \
                arg_Tinj_v, arg_Tinj_i, arg_k_v, arg_k_i = param_nearest_init(arg_mdot, arg_L2, arg_L1, arg_grad, arg_D, arg_Tinj, arg_k)

    # initial time conditions
    # arg_i = 160 # 4
    # arg_v = 40 # 20
    arg_i = 160 - (1*101) 
    arg_v = 40 - (0.25*101)

    if interp_time == "False":

        # print('FALSE')
        # print(param)
        # print(param_y.shape) # (20,)
        # print(mdot_x.shape) # (26,)
        # print(mdot_ij[:,0].shape) # (26,)
        # print(param_ij[0,:].shape) # (20,)
        
        if param == "Horizontal Extent (m)":
            arg_L2_i = slice(None)
        if param == "Vertical Extent (m)":
            arg_L1_i = slice(None)
        if param == "Geothermal Gradient (K/m)":
            arg_grad_i = slice(None)
        if param == "Borehole Diameter (m)":
            arg_D_i = slice(None)
        if param == "Injection Temperature (˚C)":
            arg_Tinj_i = slice(None)
        if param == "Rock Thermal Conductivity (W/m-K)":
            arg_k_i = slice(None)

        if fluid == "sCO2" and case == "utube":

            kWe_avg_flipped = np.transpose(u_sCO2.kWe_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i])
            kWt_avg_flipped = np.transpose(u_sCO2.kWt_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i])
            Tout_flipped = np.transpose(u_sCO2.Tout[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, arg_i])
            Pout_flipped = np.transpose(u_sCO2.Pout[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, arg_i])

        if fluid == "sCO2" and case == "coaxial":

            kWe_avg_flipped = np.transpose(c_sCO2.kWe_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i])
            kWt_avg_flipped = np.transpose(c_sCO2.kWt_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i])
            Tout_flipped = np.transpose(c_sCO2.Tout[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, arg_i])
            Pout_flipped = np.transpose(c_sCO2.Pout[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, arg_i])

        if fluid == "H2O" and case == "utube":

            kWe_avg_flipped = np.transpose(u_H2O.kWe_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i])
            kWt_avg_flipped = np.transpose(u_H2O.kWt_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i])
            Tout_flipped = np.transpose(u_H2O.Tout[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, arg_i])
            Pout_flipped = np.transpose(u_H2O.Pout[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, arg_i])

            # print(kWe_avg_flipped.shape) # (20, 26)
            # print('\n')

        if fluid == "H2O" and case == "coaxial":

            kWe_avg_flipped = np.transpose(c_H2O.kWe_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i])
            kWt_avg_flipped = np.transpose(c_H2O.kWt_avg[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i])
            Tout_flipped = np.transpose(c_H2O.Tout[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, arg_i])
            Pout_flipped = np.transpose(c_H2O.Pout[:, arg_L2_i, arg_L1_i, arg_grad_i, arg_D_i, arg_Tinj_i, arg_k_i, arg_i])

    if interp_time == "True": # would this even make sense? It doesn't because can manipulate the other factors, when they should be fixed?

        # print('TRUE')
        # print(param)
        # print(param_y.shape) # (20,)
        # print(mdot_x.shape) # (26,)
        # print(mdot_ij[:,0].shape) # (26,)
        # print(param_ij[0,:].shape) # (20,)

        point = (arg_L2, arg_L1, arg_grad, arg_D, arg_Tinj + to_kelvin_factor, arg_k, arg_v) # to kelvin
        # point2 = (arg_mdot, arg_L2, arg_L1, arg_grad, arg_D, arg_Tinj + to_kelvin_factor, arg_k) # to kelvin

        if fluid == "sCO2" and case == "utube":
            try:
                sCO2_kWe, sCO2_kWt = u_sCO2.interp_kW_contour(param, point)
                sCO2_Tout, sCO2_Pout = u_sCO2.interp_outlet_states_contour(param, point) 
                kWe_avg_flipped, kWt_avg_flipped, Tout_flipped, Pout_flipped = rename_for_contour(sCO2_kWe, sCO2_kWt, sCO2_Tout, sCO2_Pout)

            except ValueError as e:
                kWe_avg_flipped, kWt_avg_flipped, Tout_flipped, Pout_flipped = blank_data_kW() 
                error_message = parse_error_message(e=e, e_name='Err SubContour1')
                error_messages_dict['Err SubContour1'] = error_message
        
        if fluid == "sCO2" and case == "coaxial":
            try:
                sCO2_kWe, sCO2_kWt = c_sCO2.interp_kW_contour(param, point)
                sCO2_Tout, sCO2_Pout = c_sCO2.interp_outlet_states_contour(param, point)
                kWe_avg_flipped, kWt_avg_flipped, Tout_flipped, Pout_flipped = rename_for_contour(sCO2_kWe, sCO2_kWt, sCO2_Tout, sCO2_Pout)

            except ValueError as e:
                kWe_avg_flipped, kWt_avg_flipped, Tout_flipped, Pout_flipped = blank_data_kW() 
                error_message = parse_error_message(e=e, e_name='Err SubContour2')
                error_messages_dict['Err SubContour2'] = error_message
        
        if fluid == "H2O" and case == "utube":
            try:
                H2O_kWe, H2O_kWt = u_H2O.interp_kW_contour(param, point)
                H2O_Tout, H2O_Pout = u_H2O.interp_outlet_states_contour(param, point)
                kWe_avg_flipped, kWt_avg_flipped, Tout_flipped, Pout_flipped = rename_for_contour(H2O_kWe, H2O_kWt, H2O_Tout, H2O_Pout)

            except ValueError as e:
                kWe_avg_flipped, kWt_avg_flipped, Tout_flipped, Pout_flipped = blank_data_kW() 
                error_message = parse_error_message(e=e, e_name='Err SubContour3')
                error_messages_dict['Err SubContour3'] = error_message
        
            # print(kWe_avg_flipped.shape) # should have an x and y right for the z component?
            # print('\n')

        if fluid == "H2O" and case == "coaxial":
            try:
                # H2O_kWe, H2O_kWt = c_H2O.interp_kW(point2)
                H2O_kWe, H2O_kWt = c_H2O.interp_kW_contour(param, point)
                H2O_Tout, H2O_Pout = c_H2O.interp_outlet_states_contour(param, point)
                kWe_avg_flipped, kWt_avg_flipped, Tout_flipped, Pout_flipped = rename_for_contour(H2O_kWe, H2O_kWt, H2O_Tout, H2O_Pout)

            except ValueError as e:
                kWe_avg_flipped, kWt_avg_flipped, Tout_flipped, Pout_flipped = blank_data_kW() 
                error_message = parse_error_message(e=e, e_name='Err SubContour4')
                error_messages_dict['Err SubContour4'] = error_message
        

    fig = make_subplots(rows=2, cols=2, start_cell="top-left",
                        horizontal_spacing = 0.12,
                        vertical_spacing = 0.12,
                        subplot_titles=('40-Year Average Exergy (kWe)', 
                                       '40-Year Average Thermal Output (kWt)',
                                       'Outlet Temperature (°C)',
                                       'Outlet Pressure (Pa)')
           )


    contour_formatting = dict(showlabels = True, #coloring ='heatmap'
                              labelfont = dict( size = 14,
                                                color = 'white'
                                                )
                              )
    fig.add_trace( 
        go.Contour(showscale=False,
            colorbar=dict(title='Average Exergy (kWe)'), #titleside='right'
            colorscale=colorscaleR,
            name="",
            hovertemplate='<b>Mass Flow Rate (kg/s)</b>: %{x:.1f}<br><b>Y</b>: %{y:.4f}<br><b>Average Exergy (kWe)</b>: %{z:.5f} ',
            contours=contour_formatting,
            z=kWe_avg_flipped, y=param_ij[0,:], x=mdot_ij[:,0]
        ), row=1, col=1 )
                  

    fig.add_trace( 
        go.Contour(showscale=False,
            colorbar=dict(title='Average Thermal Output (kWe)'),
            colorscale=colorscaleY,
            name="",
            hovertemplate='<b>Mass Flow Rate (kg/s)</b>: %{x:.1f}<br><b>Y</b>: %{y:.4f}<br><b>Average Thermal Output (kWt)</b>: %{z:.5f} ',
            contours=contour_formatting,
            z=kWt_avg_flipped, y=param_ij[0,:], x=mdot_ij[:,0]
        ), row=1, col=2 )


    fig.add_trace( 
        go.Contour(showscale=False,
            colorbar=dict(title='Outlet Temperature (°C)'),
            colorscale=colorscaleB,
            name="",
            hovertemplate='<b>Mass Flow Rate (kg/s)</b>: %{x:.1f}<br><b>Y</b>: %{y:.4f}<br><b>Outlet Temperature (°C)</b>: %{z:.5f} ',
            contours=contour_formatting,
            z=Tout_flipped - to_kelvin_factor, y=param_ij[0,:], x=mdot_ij[:,0] # to celsius
        ), row=2, col=1 )

    fig.add_trace( 
        go.Contour(showscale=False,
            colorbar=dict(title='Outlet Pressure (Pa)'),
            colorscale=colorscaleG,
            name="",
            hovertemplate='<b>Mass Flow Rate (kg/s)</b>: %{x:.1f}<br><b>Y</b>: %{y:.4f}<br><b>Outlet Pressure (Pa)</b>: %{z:.0f} ',
            contours=contour_formatting,
            z=Pout_flipped, y=param_ij[0,:], x=mdot_ij[:,0]
        ), row=2, col=2 )

    fig = update_layout_properties_subsurface_contours(fig, param)
    
    return fig, error_messages_dict



# --------------------------------------------------------------------------------------------------------------------
# ECONOMIC PERFORMANCE PLOTS
# --------------------------------------------------------------------------------------------------------------------


def generate_econ_lineplots(interp_time, case, end_use, fluid,
                            mdot, L2, L1, grad, D, Tinj, k,
                            Drilling_cost_per_m, Discount_rate, Lifetime, 
                            Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure,
                            scale):

    # -----------------------------------------------------------------------------------------------------------------
    # Creates Plotly with 4 subplot lineplots:
    #
    #       heat prod vs. time
    #       annual heat prod vs. time
    #       electricity prod vs. time
    #       annual electricity prod vs. time
    #
    # -----------------------------------------------------------------------------------------------------------------

    lcoh_sCO2 = '-'
    lcoh_H2O = '-'
    lcoe_sCO2 = '-'
    lcoe_H2O = '-'
    
    mean_H2O_Net_HProd = '-'
    mean_H2O_Net_EProd = '-'
    mean_sCO2_Net_HProd = '-'
    mean_sCO2_Net_EProd = '-'

    econ_values_dict = {}
    error_messages_dict = {}

    fig = make_subplots(rows=2, cols=5,
                        specs=[[{'colspan': 2}, None, {'colspan': 2}, None, {"type": "table"}],
                                [{'colspan': 2}, None, {'colspan': 2}, None, {"type": "table"}]],
                        horizontal_spacing = 0.11
                        )

    if end_use == "Heating" or end_use == "All":

        if end_use == "Heating":
            is_display_legend = True
        if end_use == "All":
            is_display_legend = True

        if fluid == "sCO2" or fluid == "All":

            try:
                teaobj_sCO2 = create_teaobject(u_sCO2, u_H2O, c_sCO2, c_H2O,
                                                case, end_use, fluid, 
                                                mdot, L2, L1, grad, D, Tinj, k,
                                                Drilling_cost_per_m, Discount_rate, Lifetime, 
                                                Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure,
                                                is_heating=True)
                lcoh_sCO2 = format( teaobj_sCO2.LCOH, '.2f')

                # Heat Production 

                fig.add_trace(go.Scatter(x=teaobj_sCO2.Linear_time_distribution, y=teaobj_sCO2.Instantaneous_heat_production/1e3,
                              hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Heat Production (MWt)</b>: %{y:.3f} ',
                              line = dict(color='black', width=lw),
                              legendgroup=labels_cat[1], name=labels[1], showlegend=is_display_legend
                              ),
                              row=1, col=1)


                # Anuual Heat Production
                fig.add_trace(go.Bar(x=np.arange(1,teaobj_sCO2.Lifetime+1), y=teaobj_sCO2.Annual_heat_production/1e6,
                                    # hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Annual Heat Production (GWh)</b>: %{y:.3f} ',
                                    name=labels[1],
                                    showlegend=False, hovertemplate='<b>Annual Heat Production (GWh)</b>: %{y:.3f}'), 
                                row=1, col=3)

                fig.add_trace(go.Scatter(x=np.arange(1,teaobj_sCO2.Lifetime+1), y=teaobj_sCO2.Annual_heat_production/1e6,
                              hovertemplate='<b>Time (year)</b>: %{x:.1f}<br>',
                              line = dict(color='black', width=lw),
                              legendgroup=labels_cat[1], name=labels[1], showlegend=False
                              ),
                              row=1, col=3)

                # Table Data
                mean_sCO2_Net_HProd = round(np.mean(teaobj_sCO2.Instantaneous_heat_production/1e3),2)

                time_dict = {"Time (year)": teaobj_sCO2.Linear_time_distribution,
                             "sCO2 Heat Production (MWt)": teaobj_sCO2.Instantaneous_heat_production/1e3,
                             "sCO2 Annual Heat Production (GWh)": teaobj_sCO2.Annual_heat_production/1e6}
                econ_values_dict.update(time_dict)

            except ValueError as e:
                fig, lcoh_sCO2 = update_blank_econ2(fig=fig, nrow1=1, ncol1=1, nrow2=1, ncol2=3)
                error_message = parse_error_message(e=e, e_name='Err Econ1a')
                error_messages_dict['Err Econ1a'] = error_message
            
            except AttributeError as e:
                fig, lcoh_sCO2 = update_blank_econ2(fig=fig, nrow1=1, ncol1=1, nrow2=1, ncol2=3)
                error_message = parse_error_message(e=e, e_name='Err Econ1b')
                error_messages_dict['Err Econ1b'] = error_message

        if fluid == "H2O" or fluid == "All":

            try:
                teaobj_H2O = create_teaobject(u_sCO2, u_H2O, c_sCO2, c_H2O,
                                                case, end_use, fluid, 
                                                mdot, L2, L1, grad, D, Tinj, k,
                                                Drilling_cost_per_m, Discount_rate, Lifetime, 
                                                Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure,
                                                is_H20=True, is_heating=True)
                lcoh_H2O = format(teaobj_H2O.LCOH, '.2f')

                # Heat Production 
                fig.add_trace(go.Scatter(x=teaobj_H2O.Linear_time_distribution, y=teaobj_H2O.Instantaneous_heat_production/1e3,
                              hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Heat Production (MWt)</b>: %{y:.3f} ',
                              line = dict(color='black', width=lw, dash='dash'),
                              legendgroup=labels_cat[0], name=labels[0], showlegend=is_display_legend
                              ),
                              row=1, col=1)

                # Anuual Heat Production
                fig.add_trace(go.Bar(x=np.arange(1,teaobj_H2O.Lifetime+1), y=teaobj_H2O.Annual_heat_production/1e6,
                                    # hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Annual Heat Production (GWh)</b>: %{y:.3f} ',
                                    name=labels[0],
                                    showlegend=False, hovertemplate='<b>Annual Heat Production (GWh)</b>: %{y:.3f}'), 
                                row=1, col=3)


                fig.add_trace(go.Scatter(x=np.arange(1,teaobj_H2O.Lifetime+1), y=teaobj_H2O.Annual_heat_production/1e6,
                              hovertemplate='<b>Time (year)</b>: %{x:.1f}<br>',
                              line = dict(color='black', width=lw, dash='dash'),
                              legendgroup=labels_cat[0], name=labels[0], showlegend=False
                              ),
                              row=1, col=3)

                # Table Data
                mean_H2O_Net_HProd = round(np.mean(teaobj_H2O.Instantaneous_heat_production/1e3),2)

                time_dict = {"Time (year)": teaobj_H2O.Linear_time_distribution,
                             "H2O Heat Production (MWt)": teaobj_H2O.Instantaneous_heat_production/1e3,
                             "H2O Annual Heat Production (GWh)": teaobj_H2O.Annual_heat_production/1e6}
                econ_values_dict.update(time_dict)
            
            except ValueError as e:
                fig, lcoh_H2O = update_blank_econ2(fig=fig, nrow1=1, ncol1=1, nrow2=1, ncol2=3)
                error_message = parse_error_message(e=e, e_name='Err Econ2a')
                error_messages_dict['Err Econ2a'] = error_message
            
            except AttributeError as e:
                fig, lcoh_H2O = update_blank_econ2(fig=fig, nrow1=1, ncol1=1, nrow2=1, ncol2=3)
                error_message = parse_error_message(e=e, e_name='Err Econ2b')
                error_messages_dict['Err Econ2b'] = error_message

    if end_use == "Electricity" or end_use == "All":

        if end_use == "Electricity":
            is_display_legend = True
            row_num = 1
        else:
            is_display_legend = False
            row_num = 2

        if fluid == "sCO2" or fluid == "All":

            try:
                teaobj_sCO2 = create_teaobject(u_sCO2, u_H2O, c_sCO2, c_H2O,
                                                case, end_use, fluid, 
                                                mdot, L2, L1, grad, D, Tinj, k,
                                                Drilling_cost_per_m, Discount_rate, Lifetime, 
                                                Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure)
                lcoe_sCO2 = format( teaobj_sCO2.LCOE, '.2f')

                # Electricity 
                fig.add_trace(go.Scatter(x=teaobj_sCO2.Linear_time_distribution, y=teaobj_sCO2.Inst_Net_Electricity_production/1e3,
                              hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Electricity Production (MWe)</b>: %{y:.3f} ',
                              line = dict(color='black', width=lw),
                              legendgroup=labels_cat[1], name=labels[1], showlegend=is_display_legend
                              ),
                              row=row_num, col=1)

                
                # Annual Electricity
                fig.add_trace(go.Bar(x=np.arange(1,teaobj_sCO2.Lifetime+1), y=teaobj_sCO2.Annual_electricity_production/1e6,
                                # hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Annual Electricity Production (GWe)</b>: %{y:.3f} ',
                                name=labels[1],
                                showlegend=False, hovertemplate='<b>Annual Electricity Production (GWe)</b>: %{y:.3f}'), 
                                row=row_num, col=3)

                fig.add_trace(go.Scatter(x=np.arange(1,teaobj_sCO2.Lifetime+1), y=teaobj_sCO2.Annual_electricity_production/1e6,
                              hovertemplate='<b>Time (year)</b>: %{x:.1f}<br>',
                              line = dict(color='black', width=lw),
                              legendgroup=labels_cat[1], name=labels[1], showlegend=False
                              ),
                              row=row_num, col=3)

                # Table Data
                time_dict = {"Time (year)": teaobj_sCO2.Linear_time_distribution,
                             "sCO2 Electricity Production (MWe)": teaobj_sCO2.Inst_Net_Electricity_production/1e3,
                             "sCO2 Annual Electricity Production (GWe)": teaobj_sCO2.Annual_electricity_production/1e6}
                econ_values_dict.update(time_dict)

                if fluid == 2 and np.in1d(3000,teaobj_sCO2.error_codes) == False and np.in1d(4000,teaobj_sCO2.error_codes) == False:

                    #CO2 isobaric lines for Ts diagram (do not change data below)
                    pvectorforTS, svectorforTS, sdome, Tdome, cycleSarray, cycleTarray = CO2_isobaric_lines(teaobj=teaobj_sCO2)

                # mean_sCO2_Net_HProd = round(np.mean(teaobj_sCO2.Instantaneous_heat_production/1e3),2)
                mean_sCO2_Net_EProd = round(np.mean(teaobj_sCO2.Inst_Net_Electricity_production/1e3),2)

            except ValueError as e:
                fig, lcoe_sCO2 = update_blank_econ2(fig=fig, nrow1=row_num, ncol1=1, nrow2=row_num, ncol2=3)
                error_message = parse_error_message(e=e, e_name='Err Econ3a')
                error_messages_dict['Err Econ3a'] = error_message

            except AttributeError as e:
                fig, lcoe_sCO2 = update_blank_econ2(fig=fig, nrow1=row_num, ncol1=1, nrow2=row_num, ncol2=3)
                error_message = parse_error_message(e=e, e_name='Err Econ3b')
                error_messages_dict['Err Econ3b'] = error_message

        if fluid == "H2O" or fluid == "All":

            try:
                teaobj_H2O = create_teaobject(u_sCO2, u_H2O, c_sCO2, c_H2O,
                                                case, end_use, fluid, 
                                                mdot, L2, L1, grad, D, Tinj, k,
                                                Drilling_cost_per_m, Discount_rate, Lifetime, 
                                                Direct_use_heat_cost_per_kWth, Power_plant_cost_per_kWe, Pre_Cooling_Delta_T, Turbine_outlet_pressure,
                                                is_H20=True)
                lcoe_H2O = format(teaobj_H2O.LCOE, '.2f')

                # Electricity 
                fig.add_trace(go.Scatter(x=teaobj_H2O.Linear_time_distribution, y=teaobj_H2O.Inst_Net_Electricity_production/1e3,
                              hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Electricity Production (MWe)</b>: %{y:.3f} ',
                              line = dict(color='black', width=lw, dash='dash'),
                              legendgroup=labels_cat[0], name=labels[0], showlegend=is_display_legend
                              ),
                              row=row_num, col=1)
                
                # Annual Electricity
                fig.add_trace(go.Bar(x=np.arange(1,teaobj_H2O.Lifetime+1), y=teaobj_H2O.Annual_electricity_production/1e6,
                                # hovertemplate='<b>Time (year)</b>: %{x:.1f}<br><b>Annual Electricity Production (GWe)</b>: %{y:.3f} ',
                                name=labels[0],
                                showlegend=False, hovertemplate='<b>Annual Electricity Production (GWe)</b>: %{y:.3f}'), 
                                row=row_num, col=3)

                fig.add_trace(go.Scatter(x=np.arange(1,teaobj_H2O.Lifetime+1), y=teaobj_H2O.Annual_electricity_production/1e6,
                              hovertemplate='<b>Time (year)</b>: %{x:.1f}<br>',
                              line = dict(color='black', width=lw, dash='dash'),
                              legendgroup=labels_cat[0], name=labels[0], showlegend=False
                              ),
                              row=row_num, col=3)

                # Table Data
                time_dict = {"Time (year)": teaobj_H2O.Linear_time_distribution,
                             "H2O Electricity Production (MWe)": teaobj_H2O.Inst_Net_Electricity_production/1e3,
                             "H2O Annual Electricity Production (GWe)": teaobj_H2O.Annual_electricity_production/1e6}
                econ_values_dict.update(time_dict)

                if fluid == 2 and np.in1d(3000,teaobj_H2O.error_codes) == False and np.in1d(4000,teaobj_H2O.error_codes) == False:

                    #CO2 isobaric lines for T-S diagram (do not change data below)
                    pvectorforTS, svectorforTS, sdome, Tdome, cycleSarray, cycleTarray = CO2_isobaric_lines(teaobj=teaobj_H2O)

                    ## HERE
                    # ax3.set_title("CO2 TS-diagram", color=crisp_white)  
                    # ax3.set_xlabel("Entropy [J/kg/K]", color=crisp_white) 
                    # ax3.set_ylabel("Temperature [K]", color=crisp_white) 
                    # ax3.bar(np.arange(1,teaobj.Lifetime+1),teaobj.Annual_electricity_production/1e6)
                    # ax3.set_xticks(np.arange(0,teaobj.Lifetime+1,2),np.arange(0,teaobj.Lifetime+1,2))
                    # repeating_plot_features(ax_i=ax3)
                    # for i in range (len(pvectorforTS)):
                    #     ax3.plot(svectorforTS,tmatrix[i,:],'b-')
                    #     ax3.plot(sdome,Tdome,'r-')    
                    #     ax3.plot(cycleSarray,cycleTarray+to_kelvin_factor,'k-+') # to kelvin
                    # #plt.axis([0, 200,0,50])

                # mean_H2O_Net_HProd = round(np.mean(teaobj_H2O.Instantaneous_heat_production/1e3),2)
                mean_H2O_Net_EProd = round(np.mean(teaobj_H2O.Inst_Net_Electricity_production/1e3),2)

            except ValueError as e:
                fig, lcoe_H2O = update_blank_econ2(fig=fig, nrow1=row_num, ncol1=1, nrow2=row_num, ncol2=3)
                error_message = parse_error_message(e=e, e_name='Err Econ4a')
                error_messages_dict['Err Econ4a'] = error_message

            except AttributeError as e:
                fig, lcoe_H2O = update_blank_econ2(fig=fig, nrow1=row_num, ncol1=1, nrow2=row_num, ncol2=3)
                error_message = parse_error_message(e=e, e_name='Err Econ4b')
                error_messages_dict['Err Econ4b'] = error_message

    fig = update_layout_properties_econ_results(fig, end_use, scale)
    fig = update_lcoh_lcoe_table(fig, fluid, end_use, lcoh_sCO2, lcoh_H2O, lcoe_sCO2, lcoe_H2O) # table
    fig.update_traces(cells_font=dict(size = 13), row=1, col=5)
    fig.update_traces(cells_font=dict(size = 13), row=2, col=5)

    econ_data_dict = {'LCOH sCO2': lcoh_sCO2, 
                        'LCOH H2O': lcoh_H2O, 
                        'LCOE sCO2': lcoe_sCO2,
                        'LCOE H2O': lcoe_H2O,
                        'Mean H2O Net HProd': mean_H2O_Net_HProd,
                        'Mean H2O Net EProd': mean_H2O_Net_EProd,
                        'Mean sCO2 Net HProd': mean_sCO2_Net_HProd,
                        'Mean sCO2 Net EProd': mean_sCO2_Net_EProd}

    return fig, econ_data_dict, econ_values_dict, error_messages_dict
